package ru.dbotthepony.mc.otm.block.matter

import net.minecraft.world.level.block.EntityBlock
import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.world.item.DyeColor
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.block.entity.BlockEntity
import ru.dbotthepony.mc.otm.block.entity.matter.MatterPanelBlockEntity
import net.minecraft.world.phys.shapes.VoxelShape
import net.minecraft.world.phys.shapes.Shapes
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.phys.shapes.CollisionContext
import net.minecraft.world.item.context.BlockPlaceContext
import ru.dbotthepony.mc.otm.block.RotatableMatteryBlock
import ru.dbotthepony.mc.otm.block.addSimpleDescription
import ru.dbotthepony.mc.otm.util.get
import ru.dbotthepony.mc.otm.util.math.BlockRotation
import ru.dbotthepony.mc.otm.util.math.BlockRotationFreedom

class MatterPanelBlock(val color: DyeColor?) : RotatableMatteryBlock(DEFAULT_MACHINE_PROPERTIES), EntityBlock {
	init {
		tooltips.colored(color)
		addSimpleDescription()
	}

	override fun newBlockEntity(blockPos: BlockPos, blockState: BlockState): BlockEntity {
		return MatterPanelBlockEntity(blockPos, blockState)
	}

	private val shapes: Map<BlockState, VoxelShape>

	init {
		registerDefaultState(getStateDefinition().any().setValue(BlockRotationFreedom.DIRECTIONAL.property, BlockRotation.SOUTH))

		shapes = getShapeForEachStateMattery {
			when (it[BlockRotationFreedom.DIRECTIONAL.property].front) {
				Direction.NORTH -> Shapes.box(
					0.0,
					0.0,
					0.625,
					1.0,
					1.0,
					1.0
				)
				Direction.EAST -> Shapes.box(
					0.0,
					0.0,
					0.0,
					0.375,
					1.0,
					1.0
				)
				Direction.WEST -> Shapes.box(
					0.625,
					0.0,
					0.0,
					1.0,
					1.0,
					1.0
				)
				Direction.DOWN -> Shapes.box(
					0.0,
					0.625,
					0.0,
					1.0,
					1.0,
					1.0
				)
				Direction.UP -> Shapes.box(
					0.0,
					0.0,
					0.0,
					1.0,
					0.375,
					1.0
				)
				else -> Shapes.box(
					0.0,
					0.0,
					0.0,
					1.0,
					1.0,
					0.375
				)
			}
		}
	}

	override fun getShape(
		p_60555_: BlockState,
		p_60556_: BlockGetter,
		p_60557_: BlockPos,
		p_60558_: CollisionContext
	): VoxelShape {
		return shapes[p_60555_]!!
	}

	override fun rotationFreedom(): BlockRotationFreedom {
		return BlockRotationFreedom.DIRECTIONAL
	}

	override fun getStateForPlacement(context: BlockPlaceContext): BlockState? {
		return defaultBlockState().setValue(BlockRotationFreedom.DIRECTIONAL.property, BlockRotation.of(context.clickedFace))
	}
}
