package ru.dbotthepony.mc.otm.client.render.blockentity

import com.mojang.blaze3d.vertex.PoseStack
import com.mojang.math.Axis
import net.minecraft.client.renderer.BlockEntityWithoutLevelRenderer
import net.minecraft.client.renderer.MultiBufferSource
import net.minecraft.client.renderer.Sheets
import net.minecraft.client.renderer.blockentity.BlockEntityRenderer
import net.minecraft.client.renderer.blockentity.BlockEntityRendererProvider
import net.minecraft.client.renderer.entity.ItemRenderer
import net.minecraft.world.inventory.InventoryMenu
import net.minecraft.world.item.BlockItem
import net.minecraft.world.item.ItemDisplayContext
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.HalfTransparentBlock
import net.minecraft.world.level.block.StainedGlassPaneBlock
import net.neoforged.neoforge.capabilities.Capabilities
import net.neoforged.neoforge.client.extensions.common.IClientFluidTypeExtensions
import net.neoforged.neoforge.fluids.FluidStack
import net.neoforged.neoforge.fluids.capability.IFluidHandler
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.mc.otm.block.entity.decorative.FluidTankBlockEntity
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.render.color
import ru.dbotthepony.mc.otm.client.render.light
import ru.dbotthepony.mc.otm.client.render.normal
import ru.dbotthepony.mc.otm.client.render.overlayCoords
import ru.dbotthepony.mc.otm.client.render.uv
import ru.dbotthepony.mc.otm.client.render.vertex
import ru.dbotthepony.mc.otm.util.math.linearInterpolation
import ru.dbotthepony.mc.otm.registry.game.MBlocks

class FluidTankRenderer(private val context: BlockEntityRendererProvider.Context) : BlockEntityRenderer<FluidTankBlockEntity> {
	override fun render(
		tile: FluidTankBlockEntity,
		partialTick: Float,
		poseStack: PoseStack,
		bufferSource: MultiBufferSource,
		packedLight: Int,
		packedOverlay: Int
	) {
		renderFluidInTank(tile.fluid, tile.fluid.fluid, poseStack, bufferSource, packedLight, packedOverlay)
	}

	object FluidTankItemRenderer : BlockEntityWithoutLevelRenderer(minecraft.blockEntityRenderDispatcher, minecraft.entityModels) {
		override fun renderByItem(
			stack: ItemStack,
			displayContext: ItemDisplayContext,
			poseStack: PoseStack,
			bufferSource: MultiBufferSource,
			packedLight: Int,
			packedOverlay: Int
		) {
			val model = minecraft.modelManager.blockModelShaper.getBlockModel(MBlocks.FLUID_TANK.defaultBlockState())

			stack.getCapability(Capabilities.FluidHandler.ITEM)?.let {
				renderFluidInTank(it, it.getFluidInTank(0), poseStack, bufferSource, packedLight, packedOverlay)
			}

			val hasFoil = stack.hasFoil()
			val fabulous = displayContext == ItemDisplayContext.GUI
				|| displayContext.firstPerson()
				|| (stack.item is BlockItem && ((stack.item as BlockItem).block is HalfTransparentBlock || (stack.item as BlockItem).block is StainedGlassPaneBlock))

			for (renderPass in model.getRenderPasses(stack, fabulous)) {
				for (renderType in renderPass.getRenderTypes(stack, fabulous)) {
					val buffer = if (fabulous) ItemRenderer.getFoilBufferDirect(bufferSource, renderType, true, hasFoil) else ItemRenderer.getFoilBuffer(bufferSource, renderType, true, hasFoil)
					minecraft.itemRenderer.renderModelLists(renderPass, stack, packedLight, packedOverlay, poseStack, buffer)
				}
			}
		}
	}

	private companion object {
		val fluidTop = 13f / 16f
		val fluidBottom = 3f / 16f
		val fluidHeight = 10f / 16f
		val fluidWidth = 15.8f / 16f

		val fluidPadding = (1f - fluidWidth) * .5f

		fun renderFluidInTank(
			fluidHandler: IFluidHandler,
			fluid: FluidStack,
			poseStack: PoseStack,
			bufferSource: MultiBufferSource,
			packedLight: Int,
			packedOverlay: Int
		) {
			if (fluid.isEmpty)
				return

			val fluidLevel = fluid.amount.toFloat() / fluidHandler.getTankCapacity(0).toFloat()

			if (fluidLevel < 0.01f)
				return

			val data = IClientFluidTypeExtensions.of(fluid.fluid)
			val texture = data.stillTexture
			val sprite = minecraft.getTextureAtlas(InventoryMenu.BLOCK_ATLAS).apply(texture)
			val tint = RGBAColor.argb(data.getTintColor(fluid))
			val gas = fluid.fluid.fluidType.isLighterThanAir

			val interp = linearInterpolation(fluidLevel, sprite.v1, sprite.v0)
			val v1 = linearInterpolation(fluidBottom, sprite.v1, sprite.v0)

			val builder = bufferSource.getBuffer(Sheets.translucentCullBlockSheet())

			poseStack.pushPose()
			poseStack.translate(0f, fluidBottom + (if (gas) 1f - fluidLevel else 0f) * fluidHeight, 0f)

			for (i in 0 .. 3) {
				poseStack.pushPose()

				poseStack.translate(0.5, 0.0, 0.5)
				poseStack.rotateAround(Axis.YP.rotationDegrees(i * 90.0f), 0.0f, 1.0f, 0.0f)
				poseStack.translate(-0.5f, 0f, fluidWidth * .5f)

				val matrix = poseStack.last().pose()

				builder.vertex(matrix, fluidPadding, 0f, 0f).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u0, v1).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
				builder.vertex(matrix, fluidPadding + fluidWidth, 0f, 0f).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u1, v1).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
				builder.vertex(matrix, fluidPadding + fluidWidth, fluidLevel * fluidHeight, 0f).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u1, interp).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
				builder.vertex(matrix, fluidPadding, fluidLevel * fluidHeight, 0f).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u0, interp).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)

				poseStack.popPose()
			}

			val matrix = poseStack.last().pose()

			if (!gas && fluidLevel < 1f) { // top
				builder.vertex(matrix, fluidPadding, fluidLevel * fluidHeight, fluidPadding + fluidWidth).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u0, sprite.v1).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
				builder.vertex(matrix, fluidPadding + fluidWidth, fluidLevel * fluidHeight, fluidPadding + fluidWidth).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u1, sprite.v1).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
				builder.vertex(matrix, fluidPadding + fluidWidth, fluidLevel * fluidHeight, fluidPadding).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u1, sprite.v0).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
				builder.vertex(matrix, fluidPadding, fluidLevel * fluidHeight, fluidPadding).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u0, sprite.v0).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
			}

			if (gas && fluidLevel < 1f) { // bottom
				builder.vertex(matrix, fluidPadding, 0f, fluidPadding).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u0, sprite.v0).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
				builder.vertex(matrix, fluidPadding + fluidWidth, 0f, fluidPadding).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u1, sprite.v0).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
				builder.vertex(matrix, fluidPadding + fluidWidth, 0f, fluidPadding + fluidWidth).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u1, sprite.v1).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
				builder.vertex(matrix, fluidPadding, 0f, fluidPadding + fluidWidth).color(tint.red, tint.green, tint.blue, tint.alpha).uv(sprite.u0, sprite.v1).overlayCoords(packedOverlay).light(packedLight).normal(0.0f, 1.0f, 0.0f)
			}

			poseStack.popPose()
		}
	}
}
