package ru.dbotthepony.mc.otm.util.collect

import it.unimi.dsi.fastutil.ints.IntCollection
import it.unimi.dsi.fastutil.ints.IntIterable
import it.unimi.dsi.fastutil.ints.IntIterator
import it.unimi.dsi.fastutil.ints.IntSortedSet
import ru.dbotthepony.mc.otm.util.math.Decimal

fun IntRange.asIterable(): IntIterable {
	return IntIterable {
		val i = this@asIterable.iterator()

		object : IntIterator {
			override fun hasNext(): Boolean {
				return i.hasNext()
			}

			override fun remove() {
				throw UnsupportedOperationException()
			}

			override fun nextInt(): Int {
				return i.nextInt()
			}
		}
	}
}

fun IntCollection.ktIterator(): kotlin.collections.IntIterator {
	return object : kotlin.collections.IntIterator() {
		private val parent = this@ktIterator.intIterator()

		override fun nextInt(): Int {
			return parent.nextInt()
		}

		override fun hasNext(): Boolean {
			return parent.hasNext()
		}
	}
}

fun IntSortedSet.ktIterator(fromElement: Int): kotlin.collections.IntIterator {
	return object : kotlin.collections.IntIterator() {
		private val parent = this@ktIterator.iterator(fromElement)

		override fun nextInt(): Int {
			return parent.nextInt()
		}

		override fun hasNext(): Boolean {
			return parent.hasNext()
		}
	}
}

inline fun <T> Iterable<T>.sumOfDecimal(selector: (T) -> Decimal): Decimal {
	var result = Decimal.ZERO
	forEach { result += selector(it) }
	return result
}

inline fun <T> Iterator<T>.sumOfDecimal(selector: (T) -> Decimal): Decimal {
	var result = Decimal.ZERO
	forEach { result += selector(it) }
	return result
}
